"use strict";
const chalk_1 = require("chalk");
const Context_1 = require("../core/Context");
const deepmerge = require("deepmerge");
const formatWebpackMessages = require("react-dev-utils/formatWebpackMessages");
const WebpackDevServer = require("webpack-dev-server");
const prepareURLs = require("../utils/prepareURLs");
const log = require("../utils/log");
module.exports = async function ({ args, rootDir, eject, plugins, getBuiltInPlugins, }) {
    const command = 'start';
    const context = new Context_1.default({
        args,
        command,
        rootDir,
        plugins,
        getBuiltInPlugins,
    });
    log.verbose('OPTIONS', `${command} cliOptions: ${JSON.stringify(args, null, 2)}`);
    let serverUrl = '';
    const { applyHook, webpack } = context;
    let configArr = [];
    try {
        configArr = await context.setUp();
    }
    catch (err) {
        log.error('CONFIG', chalk_1.default.red('Failed to get config.'));
        await applyHook(`error`, { err });
        throw err;
    }
    await applyHook(`before.${command}.load`, { args });
    // eject config
    if (eject) {
        return configArr;
    }
    if (!configArr.length) {
        const errorMsg = 'No webpack config found.';
        log.warn('CONFIG', errorMsg);
        await applyHook(`error`, { err: new Error(errorMsg) });
        return;
    }
    let devServerConfig = {
        port: args.port || 3333,
        host: args.host || '0.0.0.0',
        https: args.https || false,
    };
    for (const item of configArr) {
        const { chainConfig } = item;
        const config = chainConfig.toConfig();
        if (config.devServer) {
            devServerConfig = deepmerge(devServerConfig, config.devServer);
        }
        // if --port or process.env.PORT has been set, overwrite option port
        if (process.env.USE_CLI_PORT) {
            devServerConfig.port = args.port;
        }
    }
    await applyHook(`before.${command}.run`);
    const webpackConfig = context.customWebpackConfig(configArr.map(v => v.chainConfig.toConfig()));
    let compiler;
    try {
        compiler = webpack(webpackConfig);
    }
    catch (err) {
        log.error('CONFIG', chalk_1.default.red('Failed to load webpack config.'));
        await applyHook(`error`, { err });
        throw err;
    }
    const protocol = devServerConfig.https ? 'https' : 'http';
    const urls = prepareURLs(protocol, devServerConfig.host, devServerConfig.port);
    serverUrl = urls.localUrlForBrowser;
    // typeof(stats) is webpack.compilation.MultiStats
    compiler.hooks.done.tap('compileHook', async (stats) => {
        log.info('WEBPACK', 
        // @ts-ignore
        stats.toString({
            assets: true,
            colors: true,
            chunks: false,
            entrypoints: false,
            modules: false,
        }));
        // @ts-ignore
        const json = stats.toJson({}, true);
        const messages = formatWebpackMessages(json);
        const isSuccessful = !messages.errors.length && !messages.warnings.length;
        if (isSuccessful) {
            if (stats.stats) {
                log.info('WEBPACK', 'Compiled successfully');
            }
            else {
                log.info('WEBPACK', `Compiled successfully in ${(json.time / 1000).toFixed(1)}s!`);
            }
        }
        else if (messages.errors.length) {
            console.log(messages.errors.join('\n\n'));
        }
        else if (messages.warnings.length) {
            log.warn('WEBPACK', 'Compiled with warnings.');
            console.log(messages.warnings.join('\n\n'));
        }
        await applyHook(`after.${command}.compile`, {
            url: serverUrl,
            stats,
        });
    });
    const devServer = new WebpackDevServer(compiler, devServerConfig);
    await applyHook(`before.${command}.devServer`, {
        url: serverUrl,
        devServer,
    });
    devServer.listen(devServerConfig.port, devServerConfig.host, async (err) => {
        if (err) {
            log.info('WEBPACK', chalk_1.default.red('[ERR]: Failed to start webpack dev server'));
            log.error('WEBPACK', (err.stack || err.toString()));
        }
        console.log();
        log.info('WEBPACK', chalk_1.default.green('Starting the development server at:'));
        log.info('   - Local  : ', chalk_1.default.underline.white(urls.localUrlForBrowser));
        log.info('   - Network: ', chalk_1.default.underline.white(urls.lanUrlForTerminal));
        console.log();
        await applyHook(`after.${command}.devServer`, {
            url: serverUrl,
            devServer,
            err,
        });
    });
    return devServer;
};
